from typing import TypeVar
import geopandas as gpd
from shapely.geometry import Point, Polygon, MultiPolygon, shape
from shapely import wkt, wkb
import folium
import pandas as pd
from pandas_flavor import register_dataframe_method, register_dataframe_accessor
from .connection import database_url, query
import matplotlib


from pandas import *
from geopandas import *


def create_marker_icon(color):
    # Define a function to create a custom marker icon
    return folium.Icon(
        color=color,
        icon="circle",
        prefix="fa"
    )


def marker(location, color='red', popup=None, tooltip=None):
    return folium.Marker(location=location, icon=create_marker_icon(color), popup=popup, tooltip=tooltip)


def read_csv(file, latitude='latitude', longitude='longitude', geometry=['geometry', 'geom', 'wkb_geometry'], format='', **kwargs) -> gpd.GeoDataFrame:
    csv = pd.read_csv(file, **kwargs)
    try:
        return csv.geo(latitude=latitude, longitude=longitude, geometry=geometry, format=format, **kwargs)
    except Exception as e:
        print(e)
        return csv


T = TypeVar('T')


def read_sql(sql, database=database_url(), latitude='latitude', longitude='longitude', geometry=['geometry', 'geom', 'wkb_geometry'], format='', **kwargs) -> gpd.GeoDataFrame:
    cols, results = query(sql, database)
    sql = pd.DataFrame(results, columns=[col[0] for col in cols])
    try:
        return sql.geo(latitude=latitude, longitude=longitude, geometry=geometry, format=format, **kwargs)
    except Exception as e:
        print(e)
        return sql


@register_dataframe_accessor("geo")
class GeoDataFrameAccessor():
    def __init__(self, df):
        self._df = df

    def __call__(self, *args, **kwargs):
        self._df = create_geometry_column(self._df, *args, **kwargs)
        return gpd.GeoDataFrame(self._df, crs='EPSG:4326')


def parse_geometry(g: str):
    if g.startswith('SRID'):
        g = g.split(';')[1]
        return wkt.loads(g)
    elif g.startswith('POLYGON') or g.startswith('MULTIPOLYGON') or g.startswith('POINT'):
        return wkt.loads(g)
    elif g.startswith('{'):
        return shape(g)
    else:
        return wkb.loads(g)


format_map = {
    'geojson': shape,
    'wkt': wkt.loads,
    'wkb': wkb.loads,
    '': parse_geometry
}


def create_geometry_column(df: pd.DataFrame, format='', geometry=['geometry', 'geom', 'wkb_geometry'], latitude='latitude', longitude='longitude'):
    df = df.copy()
    if len(df) == 0:
        raise Exception('No data in Dataset to create geometry')

    if type(geometry) == str and geometry in df:
        if format in format_map:
            df.loc[:, 'geometry'] = df.loc[:,
                                           geometry].apply(format_map[format])
        else:
            df.loc[:, 'geometry'] = df.loc[:, geometry]
        return df

    for col in geometry:
        if col in df:
            if format in format_map:
                df.loc[:, 'geometry'] = df.loc[:,
                                               col].apply(format_map[format])
            else:
                df.loc[:, 'geometry'] = df.loc[:, col]
            return df

    if longitude in df and latitude in df:
        df.loc[:, longitude] = df[longitude].apply(
            lambda x: '0.0' if x == '' or x is None else x)
        df.loc[:, latitude] = df[latitude].apply(
            lambda x: '0.0' if x == '' or x is None else x)
        geom = [Point(xy) for xy in zip(
            df[longitude], df[latitude])]
        df.loc[:, 'geometry'] = geom
        return df

    if 'geometry' not in df:
        raise Exception(
            'No geometry column found in Dataset. Please provide a geometry column or a latitude and longitude column')

    return df

    # # Create a GeoDataFrame from the DataFrame
    # gdf = gpd.GeoDataFrame(df, crs='EPSG:4326')
    # return gdf


@register_dataframe_method
def map_it(df: pd.DataFrame, format='', tooltip='name', geometry=['geometry', 'geom', 'wkb_geometry'], map=None, color='green', colormap='viridis', heatmap=None, opacity=0.5, zoom_level=5, center=[20.5937, 78.9629]):
    if isinstance(df, gpd.GeoDataFrame):
        gdf = df
    else:
        gdf = df.geo(format=format, geometry=geometry)

    geometry = 'geometry'

    _map = map if map is not None else folium.Map(
        location=center, zoom_start=zoom_level)
    errs = []

    # Check if heatmap_property is provided
    if heatmap is not None:
        # Retrieve the minimum and maximum values of the heatmap property
        min_value = gdf[heatmap].min()
        max_value = gdf[heatmap].max()

        normalize = matplotlib.colors.Normalize(vmin=min_value, vmax=max_value)
        # Create a colormap based on the provided colormap name
        colormap = matplotlib.colormaps.get_cmap(colormap)

        # Function to get the color based on the heatmap value
        def get_color(heatmap_value):
            return matplotlib.colors.to_hex(colormap(normalize(heatmap_value)))
    else:
        def get_color(x):
            return color

    def style_function(x):
        if heatmap is not None and heatmap in x:
            heatmap_value = x[heatmap]
            return {
                "fillColor": get_color(heatmap_value),
                "fillOpacity": opacity,
            }

        return {
            "fillColor": color,
            "fillOpacity": opacity,
        }
    # Add the points to the map
    for idx, row in gdf.iterrows():
        if not row[geometry].is_empty:

            geom = row[geometry]
            if isinstance(geom, Point):
                if tooltip in row:
                    _tooltip = row[tooltip]
                elif '{' in tooltip:
                    _tooltip = tooltip.format(**row.to_dict())
                else:
                    _tooltip = "{},{}".format(
                        row[geometry].x, row[geometry].y)
                # Set the color based on the heatmap property value
                if heatmap is not None:
                    heatmap_value = row[heatmap]
                    point_color = get_color(heatmap_value)
                else:
                    point_color = color
                marker(location=[row[geometry].y, row[geometry].x],
                       tooltip=_tooltip, color=point_color).add_to(_map)
            if isinstance(geom, Polygon) or isinstance(geom, MultiPolygon):
                if tooltip in row:
                    _tooltip = row[tooltip]
                elif '{' in tooltip:
                    _tooltip = tooltip.format(**row.to_dict())
                else:
                    _tooltip = ""

                folium.GeoJson(data=gpd.GeoSeries(
                    geom, [0] if heatmap is None else [row[heatmap]]).to_json(), tooltip=_tooltip, style_function=style_function).add_to(_map)
        else:
            errs.append(idx)

    return _map



from typing import Callable, Coroutine, List, Optional, Any, ParamSpec, TypeVar, TypeVarTuple
from atlas.boundary import Boundary, Point, Node, dataframe, geodataframe, load_boundaries
from strawberry import type, Schema, Private
from strawberry.dataloader import DataLoader
from shapely.geometry import shape
from shapely import wkb
import dabbas as db
from atlas.schema import field
from atlas.db import boundary_supabase


@type
class Pincode(Boundary):
    type: Private[str] = "pincode"


@type
class Country(Boundary):
    type: Private[str] = "country"
    name: str


@type
class State(Boundary):
    type: Private[str] = "state"

    @field
    def country(self) -> Country:
        return india

    @field
    def districts(self) -> List["District"]:
        return [muz]

    @field
    def pincodes(self) -> List["Pincode"]:
        pincodes = boundary_supabase.table('pincode').select(
            '*').eq("state_name", "Delhi").limit(10).execute().data
        return [Pincode(id='pincode:' + str(pincode['id']), pincode=pincode['pincode'], name=pincode['pincode'], _geometry=pincode['geom']) for pincode in pincodes]


@type
class District(Boundary):
    type: Private[str] = "district"

    @property
    @geodataframe
    def pincodes(self) -> List[Pincode]:
        pincodes = boundary_supabase.table('pincode').select(
            'id,pincode').eq("district_name", self.name).execute().data
        return [Pincode(_id=pincode['id'], name=pincode['pincode']) for pincode in pincodes]

    @property
    def state(self) -> "State":
        state = boundary_supabase.table('state').select(
            'id,name,district!inner(id)').eq("district.id", self._id).execute().data[0]
        return State(_id=state['id'], name=state['name'])

    @field(name="pincodes")
    async def pincodes_async(self) -> List[Pincode]:
        pincodes = boundary_supabase.table('pincode').select(
            'id,pincode').eq("districtname", self.name).execute().data
        return [Pincode(_id=pincode['id'], name=pincode['pincode']) for pincode in pincodes]

    @field(name="state")
    async def state_async(self) -> State:
        state = boundary_supabase.table('state').select(
            'id,name,district!inner(id)').eq("district.id", self._id).execute().data[0]
        return State(_id=state['id'], name=state['name'])


@type
class DataProperty(Node):
    name: str
    value: float
    node: Node
    description: Optional[str] = field(
        description='Description of the data point')


@type
class PoI(Point):
    type: Private[str] = "poi"
    name: str
    pass


class Atlas:
    def districts(country: Optional[str] = None,
                  state: Optional[str] = None) -> List[District]:
        assert state is not None, "State is required"
        return [District(_id=row['id'], name=row['name']) for row in boundary_supabase.table('district').select('id, name, state!inner(id,name)').eq(
            'state.name', state).execute().data]

    def district(name: Optional[str]) -> District:
        rows = boundary_supabase.table('district').select(
            'id,name').eq('name', name).execute().data
        if len(rows) == 0:
            raise Exception("District not found")
        district = rows[0]
        return District(_id=district['id'], name=district['name'])

    def countries() -> List[Country]:
        return [india]

    def stores(brand: str) -> List[PoI]:
        cols, results = db.query(
            "select store.id, store.name from store inner join brand on brand.name = store.l1_brand where brand.name = '{}'".format(brand))
        return [PoI(_id=row[0], name=row[1]) for row in results]

    def poi(query: str) -> List[PoI]:
        cond = ""
        # if category is not None:
        #     cond = cond + "poi.category = '{}'".format(category)
        # if subcategory is not None:
        #     cond = cond + "poi.subcategory = '{}'".format(subcategory)
        cols, results = db.query(
            "select id, name, geometry from poi {}".format(query))
        return [PoI(_id=row[0], name=row[1], _geometry=wkb.loads(row[2])) for row in results]

    def states(country: Optional[str] = None) -> List[State]:
        rows = boundary_supabase.table('state').select(
            'id', 'name').execute().data
        return [State(_id=state['id'], name=state['name']) for state in rows]

    def state(name: str) -> State:
        rows = boundary_supabase.table('state').select(
            'id', 'name').eq('name', name).execute().data
        if len(rows) == 0:
            raise Exception("State not found")
        state = rows[0]
        return State(_id=state['id'], name=state['name'])

    def pincodes(district: Optional[str] = None) -> List[Pincode]:
        pincodes = boundary_supabase.table('pincode').select(
            'id,pincode').ilike("district_name", district).execute().data
        return [Pincode(_id=pincode['id'], name=pincode['pincode']) for pincode in pincodes]

    def pincode(pincode: str) -> List[Pincode]:
        rows = boundary_supabase.table('pincode').select(
            'id,pincode').eq("pincode", pincode).execute().data

        if len(rows) == 0:
            raise Exception("Pincode not found")
        pincode = rows[0]
        return Pincode(_id=pincode['id'], name=pincode['name'])


districts = geodataframe(Atlas.districts)
district = Atlas.district
states = geodataframe(Atlas.states)
state = Atlas.state
pincodes = geodataframe(Atlas.pincodes)
countries = dataframe(Atlas.countries)
poi = geodataframe(Atlas.poi)
stores = Atlas.stores
pincode = Atlas.pincode


@type
class Query:
    @field
    def const(self, t: str) -> str:
        return t

    @field
    def query(self) -> "Query":
        return Query()

    countries = field(Atlas.countries)
    districts = field(Atlas.districts)
    district = field(Atlas.district)
    states = field(Atlas.states)
    state = field(Atlas.state)
    pincodes = field(Atlas.pincodes)
    poi = field(Atlas.poi)
    stores = field(Atlas.stores)
    pincode = field(Atlas.pincode)

    @field
    def affluence(self, location: str, segment_by: str) -> List[DataProperty]:
        return [DataProperty(id='affluence:state:1', name='affluence', value=0.5, node=bihar, description='Affluence')]


schema = Schema(query=Query)



import os
from pandas import DataFrame


def hello():
    print("Hello, World!")


def database_url():
    try:
        pg_url = os.environ['DATABASE_URL']
        return pg_url
    except KeyError:
        pg_user = os.environ.get('PG_USER', 'postgres')
        pg_password = os.environ.get('PG_PASSWORD', 'postgres')
        pg_host = os.environ.get('PG_HOST', 'localhost')
        pg_port = os.environ.get('PG_PORT', 5432)
        pg_dbname = os.environ.get('PG_DBNAME', 'postgres')
        return f'postgresql://{pg_user}:{pg_password}@{pg_host}:{pg_port}/{pg_dbname}'


def database(url = database_url()):
    from sqlalchemy import create_engine
    return create_engine(url)


def upload(df: DataFrame, name: str, database=database(), **kwargs):
    df.to_sql(name, database, **kwargs)


def connect(url=database_url()):
    import psycopg2
    conn = psycopg2.connect(url)

    return conn


def query(sql, url=database_url()):
    conn = connect(url)
    cur = conn.cursor()

    try:
        # Execute query and fetch results
        cur.execute(sql)
        results = cur.fetchall()
        cur.close()
        conn.close()
    except Exception as e:
        cur.close()
        conn.close()
        raise e
    finally:
        if cur is not None:
            cur.close()
        conn.close()

    return cur.description, results



from typing import Any, Callable, Coroutine, Generic, NewType, List, Optional, ParamSpec, TypeVar
from strawberry import type, interface, scalar, Private, enum
from strawberry.dataloader import DataLoader
from enum import Enum
from supabase import create_client
from shapely.geometry import shape
import pyproj
import psycopg2
import dabbas as db
from atlas.schema import Node, field
from atlas.db import boundary_supabase


@enum
class Unit(Enum):
    SQUARE_KILOMETERS = "square_kilometers"


JSON = scalar(
    NewType("JSON", object),
    description="The `JSON` scalar type represents JSON values as specified by ECMA-404",
    serialize=lambda v: v,
    parse_value=lambda v: v,
)

GeoJSON = scalar(
    NewType("GeoJSON", object),
    description="The `GeoJSON` scalar type represents JSON values as specified by ECMA-404",
    serialize=lambda v: v,
    parse_value=lambda v: v,
)

# Boundaries

transformer = pyproj.Transformer.from_crs('EPSG:4326', 'WGS84', always_xy=True)


@type
class Geometry(Node):
    _geometry: Private[dict] = None
    _id: Private[int]
    name: str

    @property
    def uuid(self):
        return self.id()

    @property
    async def geometry_async(self):
        if self._geometry is not None:
            return self._geometry

        raise NotImplementedError

    @property
    def geometry(self):
        if self._geometry is not None:
            return self._geometry

        raise NotImplementedError

    @property
    def geoseries(self):
        return db.GeoSeries(shape(self.geometry), crs=4326)

    def _repr_html_(self):
        return self.gdf._repr_html_()

    @property
    async def geoseries_async(self) -> db.GeoSeries:
        return db.GeoSeries(shape(await self.geometry_async), crs=4326)

    @property
    async def gdf_async(self):
        return db.GeoDataFrame({"id": self.id(), "name": self.name, "geometry": await self.geoseries_async})

    @property
    def gdf(self):
        return db.GeoDataFrame({"id": self.id(), "name": self.name, "geometry":  self.geoseries})

    def map_it(self, *args, **kwargs):
        return self.gdf.map_it(*args, **kwargs)

    def plot(self, *args, **kwargs):
        return self.gdf.plot(*args, **kwargs)

    @property
    async def utm_geoseries_async(self):
        series = await self.geoseries_async
        utm_crs = series.estimate_utm_crs()
        return series.to_crs(crs=utm_crs)

    @property
    def utm_geoseries(self):
        series = self.geoseries
        utm_crs = series.estimate_utm_crs()
        return series.to_crs(crs=utm_crs)

    @field
    def properties(self) -> "Boundary":
        return self


@type
class Point(Geometry):
    pass


@type
class Boundary(Geometry):
    _area: Private[float] = None
    _centroid: Private[GeoJSON] = None

    @property
    async def geometry_async(self):
        if self._geometry is not None:
            return self._geometry

        self._geometry = await boundary_loader.load_async(self.type, self._id)
        return self._geometry

    @property
    def geometry(self):
        if self._geometry is not None:
            return self._geometry

        self._geometry = boundary_supabase.table(self.type).select('id,geom').eq(
            'id', self._id).execute().data[0]['geom']
        return self._geometry

    @field
    async def boundary(self) -> GeoJSON:
        if await self.geometry_async is None:
            raise Exception("Geometry is not defined")

        return await self.geometry_async

    @field
    def affluence(self) -> float:
        return 0

    @field(name="centroid")
    async def centroid_async(self) -> GeoJSON:
        if self._centroid is not None:
            return self._centroid

        if await self.geometry_async is None:
            raise Exception("Geometry is not defined")

        series = await self.utm_geoseries_async
        return series.centroid.to_crs(crs=4326)[0].__geo_interface__

    @field(name="area")
    async def area_async(self, unit: Optional[Unit] = Unit.SQUARE_KILOMETERS) -> float:
        if await self.geometry_async is None:
            raise Exception("Geometry is not defined")

        series = await self.utm_geoseries_async
        area_in_sq_mts = series.area[0]
        if unit == Unit.SQUARE_KILOMETERS:
            area_in_sq_kms = area_in_sq_mts / 1000000
            return area_in_sq_kms

    @property
    def area(self) -> float:
        if self.geometry is None:
            raise Exception("Geometry is not defined")

        area_in_sq_mts = self.utm_geoseries.area[0]
        area_in_sq_kms = area_in_sq_mts / 1000000
        return area_in_sq_kms

    @field
    def count(self) -> int:
        return 0

    @field
    async def bounds(self) -> List[float]:
        return (await self.utm_geometry).bounds

    @field
    def properties(self) -> "Boundary":
        return self


def get_boundary_async(table: str) -> Callable[[List[str]], Coroutine[Any, Any, List[object]]]:
    async def loader(keys):
        print(keys)
        dic = {}
        for row in boundary_supabase.table(table).select('id,geom').in_('id', keys).order('id').execute().data:
            dic[row['id']] = row['geom']
        return [dic[key] for key in keys]

    return loader


def get_boundary(table: str) -> Callable[[List[str]], List[object]]:
    def loader(keys):
        print(keys)
        dic = {}
        for row in boundary_supabase.table(table).select('id,geom').in_('id', keys).order('id').execute().data:
            dic[row['id']] = row['geom']
        return [dic[key] for key in keys]

    return loader


class BoundaryLoader:
    loaders = {}

    def load_batch(self, type: str, ids: List[int]):
        loader = get_boundary(type)
        return loader(ids)

    async def load_async(self, type, id):
        if type not in self.loaders:
            self.loaders[type] = DataLoader(
                get_boundary_async(type), max_batch_size=100)

        return await self.loaders[type].load(id)


boundary_loader = BoundaryLoader()


def load_boundaries(type: str, items: List["Boundary"]):
    return [shape(geom) for geom in boundary_loader.load_batch(type, [item._id for item in items])]


P = ParamSpec("P")
PR = ParamSpec("PR")
T = TypeVar("T")


class BoundaryDataFrame(db.GeoDataFrame, Generic[T]):
    rows: List[T] = []

    def __init__(self, data=None, *args, geometry=None, crs=4326, **kwargs):
        if len(data) > 0 and data[0]._geometry is None:
            geometry = load_boundaries(data[0].type, data)
        else:
            geometry = [row._geometry for row in data]
        super().__init__([{"uuid": row.uuid, "name": row.name} for row in data], *args, geometry=geometry,
                         crs=crs, **kwargs)

        self.rows.extend(data)
        proj_df = self.to_crs(crs=3857)
        self['boundary'] = self['geometry']
        self['area'] = proj_df.area / 10**6
        self[['min_lon', 'min_lat', 'max_lon', 'max_lat']] = self.bounds
        self['centroid'] = proj_df.centroid.to_crs(crs=4326)
        self['obj'] = data
        if geometry == 'centroid':
            self['geometry'] = self['centroid']


def from_boundaries(data, columns=['uuid', 'name'], crs=4326, geometry='boundary', **kwargs):
    geometry = load_boundaries(data[0].type, data)
    df = db.GeoDataFrame([{"uuid": row.uuid, "name": row.name} for row in data], geometry=geometry,
                         columns=columns, crs=crs, **kwargs)
    proj_df = df.to_crs(crs=3857)
    df['boundary'] = df['geometry']
    df['area'] = proj_df.area / 10**6
    df[['min_lon', 'min_lat', 'max_lon', 'max_lat']] = df.bounds
    df['centroid'] = proj_df.centroid.to_crs(crs=4326)

    if geometry == 'centroid':
        df['geometry'] = df['centroid']
    return df


def geodataframe(fn: Callable[P, List[T]]) -> Callable[P, BoundaryDataFrame[T]]:
    def wrapped(*args, **kwargs):
        return BoundaryDataFrame(fn(*args, **kwargs))
    return wrapped


def dataframe(fn: Callable[PR, List[Any]]) -> Callable[PR, db.DataFrame]:
    def wrapped(*args, **kwargs):
        return db.DataFrame(fn(*args, **kwargs))
    return wrapped



import threading
import os
import inspect
import hashlib
import pickle


def cache(force=False, src=True, path=None):
    lock = threading.Lock()

    def decorator(func):
        # if src:
        #     cache_key = hashlib.sha256()
        #     cache_key.update(pickle.dumps(inspect.getsource(func)))
        #     cache_key = cache_key.hexdigest()
        #     cache_path = path if path is not None else '.func_cache/{}-{}'.format(
        #         func.__name__, cache_key)
        # else:
        cache_path = path if path is not None else '.func_cache/{}'.format(
            func.__name__)

        if os.path.exists(cache_path):
            with open(cache_path, 'rb') as f:
                local_cache = pickle.load(f)
        else:
            local_cache = {}

        def wrapper(*args, **kwargs):
            cache_key = hashlib.sha256()
            if src:
                cache_key.update(pickle.dumps(inspect.getsource(func)))
            else:
                cache_key.update(pickle.dumps(func.__name__))
            cache_key.update(pickle.dumps(args))
            cache_key.update(pickle.dumps(kwargs))

            cache_key = cache_key.hexdigest()

            with lock:
                if not force and cache_key in local_cache:
                    return local_cache[cache_key]

            result = func(*args, **kwargs)

            with lock:
                local_cache[cache_key] = result

                if not os.path.exists('.func_cache'):
                    os.makedirs('.func_cache')

                with open(cache_path, 'wb') as f:
                    pickle.dump(local_cache, f)

            return result

        def get_cached(*args, **kwargs):
            cache_key = hashlib.sha256()
            if src:
                cache_key.update(pickle.dumps(inspect.getsource(func)))
            else:
                cache_key.update(pickle.dumps(func.__name__))
            cache_key.update(pickle.dumps(args))
            cache_key.update(pickle.dumps(kwargs))

            cache_key = cache_key.hexdigest()

            with lock:
                if not force and cache_key in local_cache:
                    return local_cache[cache_key]

            return None

        wrapper.get_cached = get_cached
        wrapper.cache_path = cache_path
        wrapper.cache = local_cache
        wrapper.lock = lock
        return wrapper

    return decorator





import dataclasses
from typing import Callable, Literal, Mapping, List, Optional, Any, Sequence, Type, TypeVar, Union
from strawberry import BasePermission, Private, interface
from strawberry.field import StrawberryField
from strawberry.extensions import FieldExtension
from strawberry.annotation import StrawberryAnnotation


T = TypeVar("T")


class Field(StrawberryField):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __call__(self,  *args, **kwargs) -> StrawberryField:
        resolver = args[0]
        return super().__call__(resolver)


def field(
    resolver: Optional[Callable[[], T]] = None,
    *,
    name: Optional[str] = None,
    is_subscription: bool = False,
    description: Optional[str] = None,
    permission_classes: Optional[List[Type[BasePermission]]] = None,
    deprecation_reason: Optional[str] = None,
    default: Any = dataclasses.MISSING,
    default_factory: Union[Callable[..., object],
                           object] = dataclasses.MISSING,
    metadata: Optional[Mapping[Any, Any]] = None,
    directives: Optional[Sequence[object]] = (),
    extensions: Optional[List[FieldExtension]] = None,
    graphql_type: Optional[Any] = None,
    # This init parameter is used by PyRight to determine whether this field
    # is added in the constructor or not. It is not used to change
    # any behavior at the moment.
    init: Literal[True, False, None] = None,
) -> T:
    """Annotates a method or property as a GraphQL field.

    This is normally used inside a type declaration:

    >>> @strawberry.type:
    >>> class X:
    >>>     field_abc: str = strawberry.field(description="ABC")

    >>>     @strawberry.field(description="ABC")
    >>>     def field_with_resolver(self) -> str:
    >>>         return "abc"

    it can be used both as decorator and as a normal function.
    """

    type_annotation = StrawberryAnnotation.from_annotation(graphql_type)

    field_ = Field(
        python_name=None,
        graphql_name=name,
        type_annotation=type_annotation,
        description=description,
        is_subscription=is_subscription,
        permission_classes=permission_classes or [],
        deprecation_reason=deprecation_reason,
        default=default,
        default_factory=default_factory,
        metadata=metadata,
        directives=directives or (),
        extensions=extensions or [],
    )

    if resolver:
        assert init is not True, "Can't set init as True when passing a resolver."
        return field_(resolver)
    return field_

@interface
class Node:
    _id: Private[str]
    type: Private[str] = "node"

    @field
    def id(self) -> str:
        return '{}:{}'.format(self.type, self._id)

    @field
    def const(self, t: str) -> str:
        return t



from supabase import create_client
import psycopg2
import os

DB_ANON_KEY = 'eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyAgCiAgICAicm9sZSI6ICJhbm9uIiwKICAgICJpc3MiOiAic3VwYWJhc2UtZGVtbyIsCiAgICAiaWF0IjogMTY0MTc2OTIwMCwKICAgICJleHAiOiAxNzk5NTM1NjAwCn0.dc_X5iR_VP_qT0zsiyj_I_OZ2T9FtRU2BBNWN8Bu4GE'
DB_ADMIN_KEY = "eeyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJpc3MiOiJzdXBhYmFzZSIsInJlZiI6InRqbHNsYWR0dGJvbG93b29pc3JwIiwicm9sZSI6InNlcnZpY2Vfcm9sZSIsImlhdCI6MTY4MTEyMDQ5MSwiZXhwIjoxOTk2Njk2NDkxfQ.GzE799I13MX10Pbnn--h0_kSkX2md8JqCrTZyiAwniE"


def sql(query):
    try:
        conn = psycopg2.connect(database="postgres", 
                                user="reader",
                                password='password',
                                host="34.131.74.243", 
                                port="5432")
        cursor = conn.cursor()
        cursor.execute(query)
        rows = cursor.fetchall()
        cursor.close()
        conn.close()
        return rows
    except Exception as e:
        cursor.close()
        curs = conn.cursor()
        curs.execute("ROLLBACK")
        conn.commit()
        conn.close()
        raise e
    finally:
        if cursor is not None:
            cursor.close()
        conn.close()


key = os.getenv('ATLAS_KEY')

assert key is not None, "ATLAS_KEY is required"

boundary_supabase = create_client(
    'https://tjlsladttbolowooisrp.supabase.co', key)
poi_supabase = create_client('https://tjlsladttbolowooisrp.supabase.co', key)
supabase = create_client('https://tjlsladttbolowooisrp.supabase.co', key)


from setuptools import setup, find_packages

setup(
    name="dabbas",
    version="0.1.12",
    packages=find_packages(),
    entry_points={
        "console_scripts": [
            "db = db.main:main",
            "dabbas = db.main:main",
        ],
    },
    install_requires=[
        "pandas_flavor",
        "SQLAlchemy",
        "pandas",
        "geopandas",
        "psycopg2-binary",
        "shapely",
        "folium",
        "strawberry-graphql",
        "supabase"
        # Add any required dependencies here
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
)



import os
import subprocess
import sys
import argparse
import os.path
from .map import read_sql


def download(args):
    print('downloading')
    print(args.query)
    read_sql(args.query).to_csv(
        args.output_file, index=False)


def upload(args):
    # Read environment variables
    pg_user = os.environ['PG_USER']
    pg_password = os.environ['PG_PASSWORD']
    pg_host = os.environ['PG_HOST']
    pg_dbname = os.environ['PG_DBNAME']

    for input_file_and_table_name in args.input_files_and_table_names:
        if ':' in input_file_and_table_name:
            input_file, table_name = input_file_and_table_name.split(':')
        else:
            input_file = input_file_and_table_name
            table_name, _ = os.path.splitext(os.path.basename(input_file))

        # Set up ogr2ogr command
        ogr2ogr_command = [
            'ogr2ogr',
            '--config', 'OGR_STREAM_FEATURE_COUNT', '1000',
            '-f', 'PostgreSQL',
            f'PG:user={pg_user} password={pg_password} host={pg_host} dbname={pg_dbname}',
            '-nln', f'{args.schema}.{table_name}',
            input_file
        ]

        # Run the command
        subprocess.run(ogr2ogr_command, check=True)


def main(args=None):
    # Set up argument parser
    parser = argparse.ArgumentParser(
        description='Scrap is gold. Tools to play with it!')

    # Create subparsers for the "upload" and "download" commands
    subparsers = parser.add_subparsers(dest='command', required=True)
    upload_parser = subparsers.add_parser(
        'upload', help='Upload input files to a PostgreSQL database.')
    download_parser = subparsers.add_parser(
        'download', help='Download data from a PostgreSQL database using a SQL query.')

    # Add arguments for the "upload" command
    upload_parser.add_argument('input_files_and_table_names',
                               help='Input file names and optional table names (e.g., input_file1.tab[:table_name1])', nargs='+')
    upload_parser.add_argument(
        '--schema', help='Schema name to use in the database', default='raw')

    # Add arguments for the "download" command
    download_parser.add_argument('query', help='SQL query to execute')
    download_parser.add_argument(
        '--output-file', '-o', help='Output file name', default='output.csv', )

    # Parse arguments
    args = parser.parse_args(args)

    # Call the appropriate function based on the command
    if args.command == 'upload':
        # Call function to upload input files to PostgreSQL
        upload(args)
    elif args.command == 'download':
        # Call function to download data from PostgreSQL using SQL query
        download(args)


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))
